;******************************************************************************
;
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;    I (Simon Wood) can be contacted via email - unilink@mungewell.org
;    or via post - 14 Kirklands Close, Baildon, West Yorkshire, England.
;
;******************************************************************************

; Main code loop 
; Break detection is removed because there is no break when the master
; talks. The code detects IDLE mode ( data line low then high for 8ms)
; the master starts to talk only 3ms after data is low. this keeps an
; 8 ms window for gnunilink to talk with external devices.
ReceiveBreak:
	;; if BusOn signal is off
	;; we test that there is no byte transmission during 4ms
	movlw	40				; Value for 4ms (40 * 100us)
	movwf	Count

ReceiveLoop1:
	movlw	delay100us			; Value for 100us 
	movwf	Count2

ReceiveLoop2:
	btfss	OUT_PORT, BusOnBit		; Test for BusOn being high
	goto	ReceiveCommand			; if BusOn is off, I consider that it will go back on between messages.
	;; that's why it goes directly to ReceiveCommand which does wait for BusOn to go high

	btfsc	OUT_PORT, ClockBit		; Test for Clock being low
	goto	ReceiveBreak			; restart break detection if clock is high

	decfsz	Count2, 1			; Reduce Count2
	goto	ReceiveLoop2

	decfsz	Count, 1			; Reduce Count 
	goto	ReceiveLoop1

ReceiveCommand:
	; there has not been any byte transmitted since 4ms
	; so we should be between 2 packets
	; if data is high, wait until it is low, so we won't
	; poll rockbox until the next low to high transition
	; the master should start to talk only during data low phase
	; (3ms after the data low start)

	btfss	OUT_PORT, BusOnBit		; Test for BusOn being high
	goto	BusOff
	btfsc	OUT_PORT, ClockBit
	goto	ReceiveFirst			; receive packet is clock high
	btfsc	OUT_PORT, DataBit		; wait data low
	goto	ReceiveCommand

	; we are in data low state. check if there is a transmission
	; or if the data goes high again.
	; perform some init before the first clock pulse

	clrf	Checksum			; Clear checksum
	
	clrf	DataVal
	movlw	0x08
	movwf	Count

ReceiveDataLow:
	;; we test that there is no byte transmission during 4ms
	movlw	70				; Value for 7ms (70 * 100us)
	movwf	Count
	bcf	SophFlag,IdlePhase
ReceiveDataLoop1:
	movlw	delay100us			; Value for 100us 
	movwf	Count2

ReceiveDataLoop2:
	btfsc	OUT_PORT, ClockBit
	goto	ReceiveFirst			; receive packet is clock high
	btfsc	OUT_PORT, DataBit
	goto	ReceiveDataLow2High		; wait clock until data high

	decfsz	Count2, 1			; Reduce Count2
	goto	ReceiveDataLoop2

	decfsz	Count, 1			; Reduce Count 
	goto	ReceiveDataLoop1

	; here data was low during 7ms. The next data high
	; will be a real idle phase. So we can drive the slave break
	bsf	SophFlag,IdlePhase
	; wait for data going high
	goto	ReceiveDataLoop1
	

ReceiveDataLow2High:
	; low to high data transition
	; if IdlePhase is set, this means that the last data low phase was more than 7ms
	btfss	SophFlag,IdlePhase
	goto	ReceiveCommand
	; test if we must issue a break
	btfsc	SophFlag,DoBreak
	goto	ReceiveDoBreak

	IFDEF	ROCKBOX
	; test if a on pulse is occuring
	btfsc	SophFlag , DoPulse
	goto 	ReceiveDoPulse
	; test if we are in cd mode
	btfsc	SophFlag , CdMode		; if not in cd mode, do not poll
	goto	ReceiveDoPoll
	ENDIF
	goto	ReceiveCommand

ReceiveDoBreak:
	; here we issue a slave break
	; wait for 2ms
	movlw	20
	call	DoWait
	IFDEF PIC_12
	movlw	0xfe				; Enable data pin
	tris	0
	ENDIF

	IFDEF PIC_16
	bsf	STATUS, RP0			; Enable data pin as output
	bcf	OUT_ENA, DataBit
	bcf	STATUS, RP0
	ENDIF

	bcf	OUT_PORT, DataBit		; Force Data low.... 4ms

	movlw	40
	call	DoWait
	
	IFDEF PIC_12
	movlw	0xff				; Tristate data pin
	tris	0
	ENDIF

	IFDEF PIC_16
	bsf	STATUS, RP0			; Tristate data pin
	bsf	OUT_ENA, DataBit
	bcf	STATUS, RP0
	ENDIF

	; clear DoBreak ( clear dobreak when permission to talk)
	;bcf	SophFlag , DoBreak
	goto	ReceiveCommand

	IFDEF	ROCKBOX
ReceiveDoPulse:
	; a pulse if occuring, test the counter val
	decfsz	PulseCounter   ,f
	goto	ReceiveCommand			; don't poll while pulse
	decfsz	PulseCounterMsb,f
	goto	ReceiveCommand			; don't poll while pulse
	; counter null: stop the pulse
	bcf	SophFlag , DoPulse
	BANK1					; tristate pulse pin
	bsf	ON_PULSE_PIN
	BANK0
	goto	ReceiveCommand

ReceiveDoPoll:
	; test if we are in cd mode
	btfss	SophFlag , CdMode		; if not in cd mode, do not poll
	goto	ReceiveCommand

	movlw	0x40				; send rockbox poll
	call	Send
	
	; wait for a response until data goes low
ReceiveWaitResp:
	btfss	OUT_PORT, DataBit
	goto	ReceiveDataLow
	btfsc	SERIAL_PORT
	goto	ReceiveWaitResp

	call	Receive
	movwf	DataVal
	xorlw	0x01				; update time?
	btfsc	STATUS, Z
	goto	ReceiveTime
	movf	DataVal , w
	xorlw	0x02				; update disc name?
	btfsc	STATUS, Z
	goto	ReceiveDiscName
	movf	DataVal , w
	xorlw	0x03				; update track name?
	btfsc	STATUS, Z
	goto	ReceiveTrackName
	movf	DataVal , w
	xorlw	0x04				; update flags?
	btfsc	STATUS, Z
	goto	ReceiveFlags
	goto	ReceiveCommand

	; update time
ReceiveTime:
	call	Receive
	movwf	DiskNum
	call	Receive
	movwf	TrackNum
	call	Receive
	movwf	Time1
	call	Receive
	movwf	Time0
	
	movlw	0x15
	movwf	Display
	call	IssueSlaveBreak
	goto	ReceiveCommand

ReceiveTrackName:
	call	Receive
	movwf	Track0
	call	Receive
	movwf	Track1
	call	Receive
	movwf	Track2
	call	Receive
	movwf	Track3
	call	Receive
	movwf	Track4
	call	Receive
	movwf	Track5
	call	Receive
	movwf	Track6
	call	Receive
	movwf	Track7
	movlw	0x10
	movwf	Display
	call	IssueSlaveBreak
	goto	ReceiveCommand
ReceiveDiscName:
	call	Receive
	movwf	Disk0
	call	Receive
	movwf	Disk1
	call	Receive
	movwf	Disk2
	call	Receive
	movwf	Disk3
	call	Receive
	movwf	Disk4
	call	Receive
	movwf	Disk5
	call	Receive
	movwf	Disk6
	call	Receive
	movwf	Disk7
	movlw	0x12
	movwf	Display
	call	IssueSlaveBreak
	goto	ReceiveCommand
ReceiveFlags:
	call	Receive
	movwf	DFlag0
	call	Receive
	movwf	DFlag1
	call	Receive
	movwf	DFlag2
	call	Receive
	movwf	DFlag3
	movlw	0x16
	movwf	Display
	call	IssueSlaveBreak
	goto	ReceiveCommand
	ENDIF
	goto	ReceiveCommand

; The HU is going off. here we log it, and wait if there are communication
; while off. If too long, then consider that power if off.
BusOff:
	IFDEF	SERIAL_LOG
	movlw	0x50
	call	Send
	ENDIF
BusOffLoop0:
	movlw	250				;25ms wait
	movwf	Count
BusOffLoop1:
	movlw	delay100us			; Value for 100us 
	movwf	Count2
BusOffLoop2:
	btfsc	PORTA , ClockBit
	goto	BusOffLog
	btfsc	PORTA , BusOnBit
	goto	BusOffBack

	decfsz	Count2, 1			; Reduce Count2
	goto	BusOffLoop2

	decfsz	Count, 1			; Reduce Count 
	goto	BusOffLoop1
; here there has not been any byte transmitted during 25ms while BusOff
; consider it power off
	goto PowerOff

BusOffBack:
	;; busOn goes high
	IFDEF	SERIAL_LOG
	movlw	0x51
	call	Send
	ENDIF
	goto	ReceiveCommand
BusOffLog:
	call	ReadByte
	goto	BusOffLoop0
		
PowerOff:
	IFDEF SERIAL_SOFT
	movlw	0x43
	call	Send
	ENDIF

	; disable power
	bcf	POWER_CTL_PIN
	; save power by sleeping
	; watchdog will wake up
SleepLoop:
	IFDEF	SLEEP_BUG
	clrwdt	; sleep does not work on some pic (I suspect crystal)
	ELSE
	sleep
	ENDIF

	btfss	OUT_PORT, BusOnBit		; Test for BusOn being high
	goto	SleepLoop
	;; power on start here
	;; send power on pulse
	clrwdt
	bsf 	POWER_CTL_PIN
	IFDEF ROCKBOX
	bsf	ON_PULSE_PIN
	BANK1
	bcf	ON_PULSE_PIN
	bcf	POWER_CTL_PIN
	BANK0
	movlw	0xff
	movwf	PulseCounter 
	movlw	2
	movwf	PulseCounterMsb
	bsf	SophFlag , DoPulse
	ENDIF
	goto	WarmBoot2

	
;ReceiveWaitDataHigh:
;	btfss	OUT_PORT, DataBit		; wait data high
;	goto	ReceiveWaitDataHigh
;	goto	ReceiveDataLow

ReceiveFirst:
	clrf	Checksum			; Clear checksum
	
	call	ReadByte			; Get RAD
	movwf	Rad
	addwf	Checksum, 1

	btfsc	STATUS, Z			; Test for a Zero RAD
	goto	ReceiveBreak			; as you're out of sync.

	call	ReadByte			; Get TAD
	movwf	Tad
	addwf	Checksum, 1

	call	ReadByte			; Get CMD1
	movwf	Cmd1
	addwf	Checksum, 1

	call	ReadByte			; Get CMD2
	movwf	Cmd2
	addwf	Checksum, 1

ReceiveParity1:
	call	ReadByte
	movwf	Parity1
	subwf	Checksum, 0			; Test checksum

	btfss	STATUS, Z
	goto	TestReinit			; Checksum was wrong!! Still check if reinit command (nasty hack)

	btfss	Cmd1, 7				; Test for 'Short' commands
	goto	ReceiveEnd
	
	IF	0				; Ignore Medium Commands to save
	goto	ReceiveBreak			; space in code....
	ELSE
RecieveMedium:
	call	ReadByte
	movwf	Data11
	addwf	Checksum, 1

	call	ReadByte
	movwf	Data12
	addwf	Checksum, 1

	call	ReadByte
	movwf	Data13
	addwf	Checksum, 1

	call	ReadByte
	movwf	Data14
	addwf	Checksum, 1

	btfss	Cmd1, 6				; Test for 'Medium' commands
	goto	ReceiveParity2

	IF	0				; Ignore Long Commands to save
	goto	ReceiveBreak			; space in code....
	ELSE
RecieveLong:
	call	ReadByte
	movwf	Data21
	addwf	Checksum, 1

	call	ReadByte
	movwf	Data22
	addwf	Checksum, 1

	call	ReadByte
	movwf	Data23
	addwf	Checksum, 1

	call	ReadByte
	movwf	Data24
	addwf	Checksum, 1

	call	ReadByte
	movwf	Data25
	addwf	Checksum, 1
	ENDIF

ReceiveParity2:
	call	ReadByte
	movwf	Parity2
	subwf	Checksum, 0			; Test checksum

	btfss	STATUS, Z
	goto	ReceiveBreak			; Checksum was wrong!!
	ENDIF

ReceiveEnd:
	clrwdt
	call	ReadByte
	movwf	EndByte

	btfss	STATUS, Z
	goto	ReceiveBreak			; EndByte was non-zero!!

	goto	Interpret

TestReinit:
	;; this nasty hack checks that the received malformed packet is a reinit command
	movf	Parity1,w
	xorlw	0x29
	btfss	STATUS, Z
	goto	ReceiveBreak

	movf	Tad,w
	xorlw	0x10
	btfss	STATUS, Z
	goto	ReceiveBreak

	movf	Cmd1,w
	xorlw	0x01
	btfss	STATUS, Z
	goto	ReceiveBreak

	movf	Cmd2,w
	xorlw	0x00
	btfss	STATUS, Z
	goto	ReceiveBreak

	goto	WarmBoot
